from enum import Flag
from typing import List, Dict

from vpt_core.io.image import ImageSet
from vpt_plugin_watershed.watershed import key_entity_fill_channel


cell_aliases = frozenset({"cell", "cells"})
nuclei_aliases = frozenset({"nuclei", "nucleus"})


class Entity(Flag):
    Seeds = 1
    Cyto = 2


def get_detected_entity(images: ImageSet, segmentation_parameters: Dict, entities: List[str]) -> Entity:
    if len(entities) > 2 or len(entities) < 1:
        raise ValueError(f"Invalid number of output results: {len(entities)} entities were requested")
    if len(entities) > 1:
        if len(cell_aliases.intersection(entities)) > 0 and len(nuclei_aliases.intersection(entities)) > 0:
            return Entity.Seeds | Entity.Cyto
        else:
            raise ValueError(
                "The segmentation method supports two entities output only in the case of cell and nucleus entities"
            )
    if entities[0] not in nuclei_aliases:
        if len(images.keys()) > 1 and images.get(segmentation_parameters.get(key_entity_fill_channel, "")) is not None:
            return Entity.Cyto
    return Entity.Seeds



import warnings
from typing import Dict, List, Tuple, Optional

import numpy as np
from skimage import measure
from skimage.segmentation import watershed
from vpt_core import log
from vpt_core.io.image import ImageSet

from vpt_plugin_watershed.stardist.seeds import StardistResult, StardistSeedsExtractor
from vpt_plugin_watershed.watershed import key_entity_fill_channel, key_seed_channel, key_stardist_model
from vpt_plugin_watershed.watershed.entity import get_detected_entity, Entity
from vpt_plugin_watershed.watershed.seeds import prepare_watershed_images, separate_merged_seeds


def to_sd_extractor_params(params: Dict) -> Dict:
    return {
        "normalization_range": params.get("normalization_range", (1, 99.8)),
        "min_size": params.get("min_diameter", 16),
        "max_size": params.get("max_diameter", 256),
    }


def get_watershed_seeds(images: ImageSet, segmentation_parameters: Dict, entity: Entity):
    # Load the seed image
    seeds = np.array(images.as_list(segmentation_parameters.get(key_seed_channel, "")))
    if seeds.size == 0:
        raise AttributeError(
            f"{key_seed_channel} and {key_entity_fill_channel} must be specified in segmentation_parameters"
        )

    log.info("stardist initialization")
    extractor = StardistSeedsExtractor(
        segmentation_parameters[key_stardist_model], **to_sd_extractor_params(segmentation_parameters)
    )

    log.info("extracting seeds")
    nuclei, seeds = extractor.extract_seeds(seeds, entity)
    log.info(f"detected seeds on each level: {str(',').join((str(len(x)) for x in nuclei))}")

    log.info("separate_merged_seeds")
    morph_r = segmentation_parameters.get("morphology_r", 20)
    seeds = separate_merged_seeds(seeds, morph_r)
    seeds = measure.label(seeds)

    return nuclei, seeds


def run_watershed(
    images: ImageSet, segmentation_parameters: Dict, entity: List[str]
) -> Tuple[Optional[List[StardistResult]], Optional[np.ndarray]]:
    warnings.filterwarnings("ignore", message=".*deprecated and will be removed in Pillow 10.*")  # shapely
    warnings.filterwarnings("ignore", message=".*the `scipy.ndimage.morphology` namespace is deprecated.*")
    warnings.filterwarnings("ignore", message=".*the `scipy.ndimage.measurements` namespace is deprecated.*")

    result_entity = get_detected_entity(images, segmentation_parameters, entity)
    # Use the nuclear channel to derive the seeds for watershed
    nuclei, seeds = get_watershed_seeds(images, segmentation_parameters, result_entity)

    # If exporting nuceli, return early with the nuclear output
    if not Entity.Cyto & result_entity:
        return nuclei, None

    # Load the cyto images
    cyto_images = np.array(images.as_list(segmentation_parameters.get(key_entity_fill_channel, "")))

    if key_entity_fill_channel in segmentation_parameters and cyto_images.size == 0:
        raise ValueError("Unable to find images in task data for the specified fill channel")

    # If no cyto images, return early with an empty result
    if cyto_images.size == 0:
        result = np.zeros(seeds.shape, dtype=np.uint8)
        return nuclei, result

    # Preprocess cyto channel to dmap and mask images
    log.info("prepare_watershed_images")
    dmap, watershed_mask = prepare_watershed_images(cyto_images)
    seeds[np.invert(watershed_mask)] = 0

    # Run 3D watershed
    log.info("watershed")
    result = watershed(dmap, seeds, mask=watershed_mask, connectivity=np.ones((3, 3, 3)), watershed_line=True)

    log.info("watershed finished")
    if not Entity.Seeds & result_entity:
        return None, result
    return nuclei, result



from typing import Tuple

import numpy as np
from pyclustering.cluster import kmedoids
from scipy import ndimage
from skimage import filters, measure, morphology

"""
This module contains utility functions for preparing images for
watershed segmentation.
"""


def separate_merged_seeds(seedsIn: np.ndarray, morph_r: int = 20) -> np.ndarray:
    """Separate seeds that are merged in 3 dimensions but are separated
    in some 2 dimensional slices.

    Args:
        seedsIn: a 3 dimensional binary numpy array arranged as (z,x,y) where
            True indicates the pixel corresponds with a seed.
    Returns: a 3 dimensional binary numpy array of the same size as seedsIn
        indicating the positions of seeds after processing.
    """

    def create_region_image(shape, c):
        region = np.zeros(shape, dtype=np.uint8)
        for x in c.coords:
            region[x[0], x[1], x[2]] = 1
        return region

    diskStruct = morphology.disk(morph_r)

    components = measure.regionprops(measure.label(seedsIn))
    seeds = np.zeros(seedsIn.shape, dtype=np.uint8)
    for c in components:
        seedImage = create_region_image(seeds.shape, c)
        localProps = [measure.regionprops(measure.label(x)) for x in seedImage]
        seedCounts = [len(x) for x in localProps]

        if all([x < 2 for x in seedCounts]):
            goodFrames = [i for i, x in enumerate(seedCounts) if x == 1]
            goodProperties = [y for x in goodFrames for y in localProps[x]]
            seedPositions = np.round([np.median([x.centroid for x in goodProperties], axis=0)]).astype(int)
        else:
            goodFrames = [i for i, x in enumerate(seedCounts) if x > 1]
            goodProperties = [y for x in goodFrames for y in localProps[x]]
            goodCentroids = [x.centroid for x in goodProperties]
            km = kmedoids.kmedoids(
                goodCentroids, np.random.choice(np.arange(len(goodCentroids)), size=np.max(seedCounts))
            )
            km.process()
            seedPositions = np.round([goodCentroids[x] for x in km.get_medoids()]).astype(int)

        for s in seedPositions:
            for f in goodFrames:
                seeds[f, s[0], s[1]] = 1

    seeds = ndimage.binary_dilation(seeds, structure=ndimage.generate_binary_structure(3, 1))
    seeds = np.array([ndimage.binary_dilation(x, structure=diskStruct) for x in seeds])

    return seeds


def prepare_watershed_images(watershedImageStack: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Prepare the given images as the input image for watershedding.

    A watershed mask is determined using an adaptive threshold and the watershed
    images are inverted so the largest values in the watershed images become
    minima and then the image stack is normalized to have values between 0
    and 1.

    Args:
        watershedImageStack: a 3 dimensional numpy array containing the images
            arranged as (z, x, y).
    Returns: a tuple containing the normalized watershed images and the
        calculated watershed mask
    """
    filterSize = int(2 * np.floor(watershedImageStack.shape[1] / 16) + 1)

    watershedMask = np.array(
        [
            ndimage.binary_fill_holes(x > filters.threshold_local(x, filterSize, method="mean", mode="wrap"))
            for x in watershedImageStack
        ]
    )

    img_min, img_max = np.min(watershedImageStack), np.max(watershedImageStack)

    if img_max > img_min:
        normalizedWatershed = 1 - (watershedImageStack - img_min) / (img_max - img_min)
    else:
        normalizedWatershed = 1 - watershedImageStack

    normalizedWatershed[np.invert(watershedMask)] = 1

    return normalizedWatershed, watershedMask



from typing import Dict, Iterable, List, Optional, Union

import pandas as pd
from vpt_core.io.image import ImageSet
from vpt_core.segmentation.polygon_utils import generate_polygons_from_mask
from vpt_core.segmentation.seg_result import SegmentationResult
from vpt_core.segmentation.segmentation_base import SegmentationBase

from vpt_plugin_watershed.stardist.utils import polygons_from_stardist
from vpt_plugin_watershed.watershed.entity import cell_aliases, nuclei_aliases
from vpt_plugin_watershed.watershed.segmentation import run_watershed
from vpt_plugin_watershed.watershed.validate import validate_task


class SegmentationMethod(SegmentationBase):
    @staticmethod
    def run_segmentation(
        segmentation_properties: Dict,
        segmentation_parameters: Dict,
        polygon_parameters: Dict,
        result: List[str],
        images: Optional[ImageSet] = None,
        transcripts: Optional[pd.DataFrame] = None,
    ) -> Union[SegmentationResult, Iterable[SegmentationResult]]:
        if images is None:
            raise ValueError("Unable to find input channels data")

        out = []
        nucleus, cells = run_watershed(images, segmentation_parameters, result)
        if nucleus is not None:
            out.append(polygons_from_stardist(nucleus, polygon_parameters))

        if cells is not None:
            out.append(generate_polygons_from_mask(cells, polygon_parameters))

        if len(out) > 1:
            if result[0] in cell_aliases and result[1] in nuclei_aliases:
                return out[::-1]
            else:
                return out

        return out[0] if len(out) == 1 else SegmentationResult()

    @staticmethod
    def validate_task(task: Dict) -> Dict:
        return validate_task(task)



import os
from typing import List, Tuple

import cv2
import numpy as np
from vpt_core.io.vzgfs import filesystem_path_split

from vpt_plugin_watershed.stardist import StardistResult
from vpt_plugin_watershed.watershed.entity import Entity

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
from vpt_core.io.output_tools import suppress_output  # noqa

from vpt_plugin_watershed.stardist import StarPolygon  # noqa


def normalize(image: np.ndarray, pmin: float, pmax: float):
    if image[image > 0].size > 0:
        [mi, ma] = np.percentile(image[image > 0], [pmin, pmax], axis=None)
        return (image.astype(np.float32) - mi) / (ma - mi + 1e-20)
    else:
        return image.astype(np.float32)


class StardistSeedsExtractor:
    def __init__(
        self,
        model_path: str,
        normalization_range: Tuple[float, float] = (1, 99.8),
        min_size: int = 16,
        max_size: int = 256,
    ):
        with suppress_output():
            from stardist.models import StarDist2D

            fs, path = filesystem_path_split(model_path)
            assert "file" in fs.protocol  # Stardist only supports local filesystem
            path_parts = path.split(fs.sep)
            parent_path, model_name = fs.sep.join(path_parts[:-1]), path_parts[-1]

            self.model = StarDist2D(None, model_name, parent_path)
            if not self.model:
                raise ValueError("Invalid stardist model")

        self.min_size = min_size
        self.max_size = max_size
        self.nrm_range = normalization_range
        self.stars = None

    def run_prediction(self, image) -> StardistResult:
        normalized_image = normalize(image, self.nrm_range[0], self.nrm_range[1])
        with suppress_output():
            n_tiles = None if image.shape[0] < 1024 else (2, 2)
            _, details = self.model.predict_instances(normalized_image, n_tiles=n_tiles)
        points, center = details["coord"], details["points"]
        result: List[StarPolygon] = []
        # filter results
        for i in range(len(center)):
            box = [[np.min(points[i][0]), np.max(points[i][0])], [np.min(points[i][1]), np.max(points[i][1])]]
            h = box[0][1] - box[0][0]
            w = box[1][1] - box[1][0]
            if h > image.shape[0] or w > image.shape[1]:
                # invalid image, return nothing
                return []

            # filter by bbox size
            if h > self.min_size and h < self.max_size and w > self.min_size and w < self.max_size:
                result.append(StarPolygon(center=center[i], points=points[i]))

        return result

    @staticmethod
    def result_to_mask(shape, results: StardistResult, star_scale: float = 0.6):
        target = np.zeros(shape, dtype=np.uint8)

        for star in results:
            triangles = []
            center = star.center
            points = star.points

            c = np.array([center[1], center[0]])
            for j in range(len(points[0]) >> 1):  # half number of points, for speed
                i1 = j + j
                i2 = (i1 + 2) % len(points[0])
                p1 = np.array([points[1][i1], points[0][i1]]) - c
                p2 = np.array([points[1][i2], points[0][i2]]) - c
                triangles.append([p1 * star_scale + c, p2 * star_scale + c, c])
            cv2.fillPoly(target, np.array(triangles, dtype=np.int32), (255, 255, 255))
        return target != 0

    def extract_seeds(self, seed_images: np.ndarray, entity: Entity) -> Tuple[List[StardistResult], np.ndarray]:
        vector = [self.run_prediction(x) for x in seed_images]
        shape = seed_images[0].shape
        rastr = np.array([StardistSeedsExtractor.result_to_mask(shape, x) for x in vector])

        # If the nuclear polygons are not needed, clear the data
        # A dummy array is created for logging
        if not Entity.Seeds & entity:
            vector = [[StarPolygon(center=np.array(0), points=np.array(0))] * len(x) for x in vector]

        return vector, rastr



from typing import Dict, List

import numpy as np
from shapely.geometry import MultiPolygon, Polygon

from vpt_core.segmentation.geometry_utils import convert_to_multipoly, get_valid_geometry
from vpt_core.segmentation.polygon_utils import PolygonCreationParameters, smooth_and_simplify
from vpt_core.segmentation.seg_result import SegmentationResult

from vpt_plugin_watershed.stardist import StardistResult


def polygons_from_stardist(sd_results: List[StardistResult], polygon_parameters: Dict) -> SegmentationResult:
    parameters = PolygonCreationParameters(**polygon_parameters)
    polys_data = []
    for z in range(len(sd_results)):
        for idx, star in enumerate(sd_results[z]):
            p = Polygon(np.fliplr(np.swapaxes(star.points, 0, 1)))
            p = smooth_and_simplify(p, parameters.smoothing_radius, parameters.simplification_tol)
            if p.is_empty:
                continue
            polys_data.append(
                {
                    SegmentationResult.detection_id_field: idx + 1,
                    SegmentationResult.cell_id_field: idx,
                    SegmentationResult.z_index_field: z,
                    SegmentationResult.geometry_field: convert_to_multipoly(get_valid_geometry(MultiPolygon([p]))),
                }
            )

    seg_result = SegmentationResult(list_data=polys_data)
    seg_result.remove_polys(lambda poly: poly.area < parameters.minimum_final_area)
    return seg_result



from typing import Dict

from vpt_core.io.output_tools import suppress_output
from vpt_core.io.vzgfs import filesystem_path_split

from vpt_plugin_watershed.watershed import key_entity_fill_channel, key_seed_channel


def watershed_model_names_to_paths(task_raw: Dict) -> Dict:
    ret = task_raw.copy()

    stardist_model = ret["segmentation_parameters"]["stardist_model"]

    fs, path = filesystem_path_split(stardist_model)
    if not fs.exists(path):
        from csbdeep.models.pretrained import get_model_folder
        from stardist.models import StarDist2D

        with suppress_output():
            model_path = get_model_folder(StarDist2D, stardist_model)
            ret["segmentation_parameters"]["stardist_model"] = model_path.as_posix()

    return ret


def validate_task(task: Dict) -> Dict:
    seed_channel = task["segmentation_parameters"].get(key_seed_channel, None)
    fill_channel = task["segmentation_parameters"].get(key_entity_fill_channel, None)

    channels = [input_data["image_channel"] for input_data in task["task_input_data"]]

    entity_types_detected = task.get("entity_types_detected", list())
    entity_list_valid = entity_types_detected and len(entity_types_detected) > 0
    if not entity_list_valid:
        raise ValueError("At least one entity type must be specified for each task")

    if seed_channel and seed_channel not in channels:
        raise ValueError(f"{seed_channel} is not in input channels")

    if fill_channel and fill_channel not in channels:
        raise ValueError(f"{fill_channel} is not in input channels")
    return watershed_model_names_to_paths(task)
